#include "BinarySearchTree.h"

BinarySearchTree::BinarySearchTree() { root_ = nullptr; }

BinarySearchTree::~BinarySearchTree() {
  if (root_)
    delete root_;
}

bool BinarySearchTree::add(int key, int type, unsigned int len, void *data) {
  Node *newNode = new Node(key, type, len, data);
  Node *parent;

  bool found = findNode(key, nullptr, &parent);

  if (found) {
    return false;
  }

  if (!parent) {
    root_ = newNode;
  } else if ((*parent) > key) {
    parent->setLeftChild(newNode);
  } else {
    parent->setRightChild(newNode);
  }

  return true;
}

bool BinarySearchTree::write(int key, int type, unsigned int len, void *data) {
  Node *node;
  bool found = findNode(key, &node, nullptr);

  if (found) {
    if (len > 0)
      node->setData(type, data);
    else {
      node->setData(-1, (void *)"NO DATA SPECIFIED!");
    }
  }

  return found;
}

bool BinarySearchTree::read(int key, int *type, unsigned int *len, void *data) {
  Node *node;

  bool found = findNode(key, &node, nullptr);

  if (found) {
    node->getData(type, len, data);
  } else {
    *type = -1;
    *len = 0;
  }

  return found;
}

bool BinarySearchTree::remove(int key) {
  Node *parent = nullptr;
  Node *node = root_;
  Node *child = nullptr;

  bool found = findNode(key, &node, &parent);

  if (!found) {
    return false;
  }

  if (node->hasLeftChild() && node->hasRightChild()) {
    Node *predecessor = node->getLeftChild();
    Node *predecessorParent = node;

    while (predecessor->hasRightChild()) {
      predecessorParent = predecessor;
      predecessor = predecessor->getRightChild();
    }

    if (predecessorParent != node) {
      predecessorParent->setRightChild(predecessor->getLeftChild());
      predecessor->setLeftChild(node->getLeftChild());
    }

    predecessor->setRightChild(node->getRightChild());

    if (!parent) {
      root_ = predecessor;
    } else if (parent->getLeftChild() == node) {
      parent->setLeftChild(predecessor);
    } else {
      parent->setRightChild(predecessor);
    }
  } else {
    if (node->hasLeftChild()) {
      child = node->getLeftChild();
    } else if (node->hasRightChild()) {
      child = node->getRightChild();
    }

    if (!parent) {
      root_ = child;
    } else if (parent->getLeftChild() == node) {
      parent->setLeftChild(child);
    } else {
      parent->setRightChild(child);
    }
  }

  return true;
}

std::vector<int> BinarySearchTree::find(unsigned int len, void *data) {
  std::vector<int> keys;

  getNodeWithDataKey(root_, &keys, len, data);

  return keys;
}

std::vector<int> BinarySearchTree::getKeys() {
  std::vector<int> keys;

  getNodeKey(root_, &keys);

  return keys;
}

void BinarySearchTree::printKeys() {
  printNode(root_);
  std::cout << std::endl;
}

void BinarySearchTree::getNodeKey(Node *node, std::vector<int> *keys) {
  if (!node)
    return;

  getNodeKey(node->getLeftChild(), keys);
  keys->push_back(node->getKey());
  getNodeKey(node->getRightChild(), keys);
}

void BinarySearchTree::getNodeWithDataKey(Node *node, std::vector<int> *keys,
                                          unsigned int len, void *data) {
  if (!node)
    return;

  char *nodeData;
  unsigned int nodeDataLen;

  getNodeWithDataKey(node->getLeftChild(), keys, len, data);

  node->getData(nullptr, &nodeDataLen, nullptr);

  if (nodeDataLen == len) {
    nodeData = new char[nodeDataLen];

    node->getData(nullptr, nullptr, nodeData);

    if (std::memcmp(data, nodeData, nodeDataLen) == 0) {
      keys->push_back(node->getKey());
    }
  }

  getNodeWithDataKey(node->getRightChild(), keys, len, data);
}

void BinarySearchTree::printNode(Node *node) {
  if (!node)
    return;

  printNode(node->getLeftChild());
  std::cout << node->getKey() << "   ";
  printNode(node->getRightChild());
}

bool BinarySearchTree::findNode(int key, Node **node, Node **parent) {
  Node *tmp = root_;
  Node *p = nullptr;

  while (tmp) {

    if (*tmp == key) {
      if (node) {
        *node = tmp;
      }
      if (parent) {
        *parent = p;
      }
      return true;
    }

    if (*tmp > key) {
      p = tmp;
      tmp = tmp->getLeftChild();
    } else {
      p = tmp;
      tmp = tmp->getRightChild();
    }
  }

  if (node) {
    *node = tmp;
  }

  if (parent) {
    *parent = p;
  }

  return false;
}