#include <fstream>
#include <iostream>
#include <sstream>

#include "BinarySearchTree.h"
#include "DataTypes.h"

BinarySearchTree *createTree(std::string filename);
void addKey(BinarySearchTree *bst, int key, std::string str);
void addKey(BinarySearchTree *bst, int key, DateStruct date);
void readKey(BinarySearchTree *bst, int key);
bool saveData(BinarySearchTree *bst, std::string filename);

int main() {
  BinarySearchTree *bst = createTree("resources/input_data.bst");

  if (!bst) {
    std::cerr << "Cannot open input file!\n";
    return 1;
  }

  std::string nameToSearch = "Sara";

  std::vector<int> keys = bst->find(nameToSearch.length(), &nameToSearch[0]);

  for (auto key : keys) {
    readKey(bst, key);
  }

  for (auto key : keys) {
    bst->remove(key);
  }

  for (auto key : keys) {
    readKey(bst, key);
  }
  
  delete bst;

  return 0;
}

BinarySearchTree *createTree(std::string filename) {
  BinarySearchTree *bst;

  std::ifstream file(filename);

  if (!file.is_open()) {
    return nullptr;
  }

  bst = new BinarySearchTree();

  int key;
  char type;
  char c;
  DateStruct date;
  std::string str;

  std::string line;
  while (std::getline(file, line)) {
    std::istringstream iss(line);

    iss >> key >> type;

    switch (type) {
    case 'D':
      iss >> date.day >> c >> date.month >> c >> date.year;
      addKey(bst, key, date);
      break;

    case 'S':
      iss >> str;
      addKey(bst, key, str);
      break;

    default:
      std::cout << "Wrong data type in line: " << line << std::endl;
    }
  }
  return bst;
}

void addKey(BinarySearchTree *bst, int key, std::string str) {
  char *data = new char[str.length()];

  std::memcpy(data, &str[0], str.length());

  bst->add(key, NODE_TYPE::STRING, str.length(), data);
}

void addKey(BinarySearchTree *bst, int key, DateStruct date) {
  char *data = new char[sizeof(DateStruct)];

  std::memcpy(data, &date, sizeof(DateStruct));

  bst->add(key, NODE_TYPE::DATE, sizeof(DateStruct), data);
}

void readKey(BinarySearchTree *bst, int key) {
  bool state = false;
  int type = 0;
  unsigned int len = 0;

  state = bst->read(key, &type, &len, nullptr);

  if (!state) {
    std::cout << "Key " << key << " not found" << std::endl;
  } else if (type == NODE_TYPE::DATE) {
    DateStruct dateStruct;

    state = bst->read(key, &type, &len, (void *)&dateStruct);

    if (len != sizeof(DateStruct)) {
      std::cout << "Invalid date data (Key: " << key << ")" << std::endl;
    } else {
      std::cout << "Key " << key << " holds date: " << dateStruct.day << "/"
                << dateStruct.month << "/" << dateStruct.year << std::endl;
    }
  } else if (type == NODE_TYPE::STRING) {
    char *str = new char[len];

    state = bst->read(key, &type, &len, str);

    std::cout << "Key " << key << " holds string of length " << len << ": \""
              << str << "\"" << std::endl;
  } else {
    std::cout << "Unknown data format (Key: " << key << ")" << std::endl;
  }
}

bool saveData(BinarySearchTree *bst, std::string filename) {
  std::ofstream file(filename);

  if (!file.is_open())
    return false;

  std::vector<int> keys = bst->getKeys();
  int type = 0;
  unsigned int len = 0;

  for (auto key : keys) {
    bst->read(key, &type, &len, nullptr);

    if (type == NODE_TYPE::DATE) {
      DateStruct dateStruct;

      bst->read(key, &type, &len, (void *)&dateStruct);

      if (len == sizeof(DateStruct)) {
        file << key << " D " << dateStruct.day << "/" << dateStruct.month << "/"
             << dateStruct.year << std::endl;
      }
    } else if (type == NODE_TYPE::STRING) {
      char *str = new char[len];

      bst->read(key, &type, &len, str);

      file << key << " S " << str << std::endl;
    }
  }

  file.close();
  return 0;
}